import logging
from os.path import abspath
from urllib.parse import unquote, urlparse
from pyls_jsonrpc.dispatchers import MethodDispatcher
from pyls_jsonrpc.endpoint import Endpoint
from pyls_jsonrpc.streams import JsonRpcStreamReader, JsonRpcStreamWriter
from mrols.analyser import MROAnalyser


log = logging.getLogger(__name__)


class MROLanguageServer(MethodDispatcher):
    """
    Implement a JSON RPC method dispatcher for the language server protocol.
    """

    MAX_WORKERS = 64

    SYNC_FULL = 1  # TextDocumentSyncKind.Full = 1
    SYNC_INCREMENTAL = 2  # TextDocumentSyncKind.Incremental = 2

    def __init__(self, rx, tx, check_parent_process=False, sync_kind=0):
        self._jsonrpc_stream_reader = JsonRpcStreamReader(rx)
        self._jsonrpc_stream_writer = JsonRpcStreamWriter(tx)
        self._check_parent_process = check_parent_process
        self._endpoint = Endpoint(self,
                                  self._jsonrpc_stream_writer.write,
                                  max_workers=self.MAX_WORKERS)
        # the synchronisation kind between the language server and client
        self._sync_kind = sync_kind if sync_kind else self.SYNC_INCREMENTAL
        # the actual MRO analyser
        self._analyser = None

    def start(self):
        """Entry point for the server."""
        # start to listen on the stream input, the consume() will write the
        # result into the stream output via the stream writer
        self._jsonrpc_stream_reader.listen(self._endpoint.consume)

    def m_initialize(self, rootUri=None, rootPath=None, **kwargs):
        """Initialise the language server and reply the capabilities."""
        root_dir = self.uri_to_abs_path(rootUri) if rootUri else rootPath
        self._analyser = MROAnalyser(root_dir=root_dir)
        log.info("Initialising custom Python MRO language server.")
        return {"capabilities": {
            "textDocumentSync": {
                "openClose": True,
                "change": self._sync_kind,
            },
            "codeLensProvider": {
                "resolveProvider": True,
            },
            "hoverProvider": True,
        }}

    def m_text_document__did_open(self, textDocument=None):
        """Split the lines of the newly opened document and store them."""
        self._analyser.replace_script_content(
            self.uri_to_abs_path(textDocument['uri']),
            textDocument['text']
        )

    def m_text_document__did_change(self,
                                    textDocument=None,
                                    contentChanges=None):
        """Update the stored document contents according to the sync kind."""
        if not contentChanges:
            return
        if self._sync_kind == self.SYNC_FULL:
            self._analyser.replace_script_content(
                self.uri_to_abs_path(textDocument['uri']),
                contentChanges[0]['text']
            )
        elif self._sync_kind == self.SYNC_INCREMENTAL:
            for change in contentChanges:
                self._analyser.update_script_content(
                    self.uri_to_abs_path(textDocument['uri']),
                    (change['range']['start']['line'],
                     change['range']['start']['character']),
                    (change['range']['end']['line'],
                     change['range']['end']['character']), change['text'])

    def m_text_document__hover(self,
                               textDocument=None,
                               position=None,
                               **_kwargs):
        """Calculate and return the hover result."""
        return self._analyser.update_fetch_hover(
            self.uri_to_abs_path(textDocument['uri']),
            (position['line'], position['character'])
        )

    def m_text_document__code_lens(self, textDocument=None, **_kwargs):
        """Calculate and return the code lens list."""
        return self._analyser.update_fetch_code_lens(
            self.uri_to_abs_path(textDocument['uri'])
        )

    def m_code_lens__resolve(self, **codeLens):
        """Fill up the details of a code lens."""
        codeLens['command'] = {
            'command': 'pythonMRO.showMRO',
            'title': 'Show MRO list',
            'arguments': [
                '\n'.join(codeLens['data']),
            ]
        }
        return codeLens

    def m_shutdown(self):
        log.info('Shutting down the service')

    def m_initialized(self):
        log.info('Service initialised')

    def m_exit(self):
        log.info('Exiting...')

    @staticmethod
    def uri_to_abs_path(uri: str) -> str:
        """
        Convert an URI to a file path.

        Args:
            uri: the URI to convert

        Returns:
            The file path of the URI
        """
        return abspath(unquote(urlparse(uri).path))



"""
This module provides examples for tests to check if a class declaration with a
lot of unnecessary spaces, new lines, change-of-line or comments between
keywords or brackets can be successfully detected by the language server.
"""
from abc import ABC


class TwistedDeclaration\
									\
									(  # strange change line and new line
										ABC\
											
											# strange comment
)\
	:

	# unnecessary space

	def\
			__init__\
				(
					\
					
						


		self
		\
			\
) -> \
	None\
:
			self\
				.\
		val = \
			\
(
			1,2,       \
	3
	,
\
		)
			
			val2\
						=\
								[
									# strange comments
									'car'      ,

									"""
						


									""",

f''
								]


import sys
from os import path


CURR_EX_ROOT_DIR = path.abspath(path.join(path.dirname(__file__), '../..'))
if CURR_EX_ROOT_DIR not in sys.path:
	sys.path.insert(0, CURR_EX_ROOT_DIR)


from class_def_in_multi_files.level_one.intermediate_defs import A
from ..level_one.intermediate_defs import D


class K3(D, A): pass


"""
This module provies examples for tests to check if a class declaration in
docstring, end-of-line comment, multi-line comment or string literals are
correctly ignored by the language server.
"""
from abc import ABC, abstractmethod  # class Ignore: pass


class DefinitionInComments(
	ABC  # class Ignore(ABC): pass
):
	"""The following class declaration should be ignored:

	class Ignore(ABC):
		
		def __init__(self):
			self.val = 1
	
	The above class declaration should be ignored.
	"""

	@abstractmethod  # class Ignore: pass
	def method(self):
		"""
		class Ignore(ABC):

			@property
			def val(self):
				return 1
		"""
		# the above class declaration should be ignored
		class_def = 'class Ignore: pass'
		return class_def  # class Ignore: pass
	
	STATIC_VAL = """

class Ignore(ABC):

	def should_ignore(self):
		return True

"""


"""
This module provides examples for tests to check if the word class in docstring,
end-of-line comment, multi-line comment or string literals are correctly ignored
by the language server.

Unnecessary class word, ABC, WordInComments
"""
from abc import ABC, abstractmethod  # unnecessary class word, ABC


class WordInComments(ABC):  # unnecessary class word, ABC, WordInComments
	"""
	This class, WordInComments, is an example class definition for testing if
	the class word and name can be correctly ignored by the language server.
	"""

	STATIC_VAL = 'class word, ABC, WordInComments'
	"""Unnecessary class word, ABC, WordInComments"""

	@property  # unnecessary class word, ABC, WordInComments
	@abstractmethod  # unnecessary class word, ABC, WordInComments
	def property_1(self):  # unnecessary class word, ABC, WordInComments
		"""Unnecessary class word, ABC, WordInComments"""
		# unnecessary class word, ABC, WordInComments
		pass  # unnecessary class word, ABC, WordInComments

	@staticmethod  # unnecessary class word, ABC, WordInComments
	def static_1():  # unnecessary class word, ABC, WordInComments
		pass

	@classmethod  # unnecessary class word, ABC, WordInComments
	def class_1(cls):  # unnecessary class word, ABC, WordInComments
		pass

	@classmethod
	def class_2(_class):  # unnecessary class word, ABC, WordInComments
		"""
		Unnecessary class word,
		ABC, WordInComments
		"""
		pass

	def method_1(
		self,  # unnecessary class word, ABC, WordInComments
		val_1: \
			str,  # unnecessary class word, ABC, WordInComments
		val_2: int,
	):  # unnecessary class word, ABC, WordInComments
		pass


class MultiLineDef(  # unnecessary class word, ABC, WordInComments
	WordInComments,  # unnecessary class word, ABC, WordInComments
):  # unnecessary class word, ABC, WordInComments
	pass


class DefInComments(
	MultiLineDef,  # class TestDef: pass
):  # class TestDef: pass
	"""
	class TestDef:
		pass
	"""

	VAL = "class TestDef: pass"
	"""class TestDef: pass"""

"""
Unnecessary multi-line at the end of file with out a new empty line for ending.
This aims to test if the language server can deal with this non-recommended end.
"""


import pytest
import yaml
import glob
from os import path
from random import randint
from typing import Callable, Generator, Sequence, Tuple, List
from mrols.analyser import MROAnalyser
from mrols.calculator import MROCalculator


class TestUtils:
	"""This class encapsulates the common utility functions for testing."""

	TEST_FILE_ROOT = path.abspath(path.dirname(__file__))
	"""The root directory of the pytest files."""
	TEST_FILE_ROOT_CHECKER = lambda f: path.isfile(f) \
				and path.basename(f).startswith('test_') \
				and (path.splitext(f)[1] == '.py')
	"""The checker to test if a file is a test file."""

	EXAMPLE_FILE_ROOT = path.abspath(path.join(TEST_FILE_ROOT, '..', 'test_data', 'examples'))
	"""The root directory of the example python files."""
	EXAMPLE_FILE_ROOT_CHECKER = lambda f: path.isfile(f) \
				and (path.splitext(f)[1] == '.py')
	"""The checker to test if a file is an example file."""

	RESULT_FILE_ROOT = path.abspath(path.join(TEST_FILE_ROOT, '..', 'test_data', 'results'))
	"""The root directory of the yaml files."""
	RESULT_FILE_ROOT_CHECKER = lambda f: path.isfile(f) \
				and (path.splitext(f)[1] in ['.yml', '.yaml'])
	"""The checker to test if a file is a yaml file."""

	@staticmethod
	def get_target_files(dir: str, ext: str = None):
		"""
		Get the target files in a directory recursively by ignoring the
		unimportant files.

		Args:
			dir: the root directory to search
			ext: the specific extension to search
		
		Returns:
			A set of the target files.
		"""
		if not ext:
			ext = ''
		candidates = set(
			glob.iglob(
				path.join(dir, '**/[!_]*' + ext), recursive=True
			)
		) - set(
			glob.iglob(
				path.join(dir, '**/__pycache__/*'), recursive=True
			)
		)
		return [c for c in candidates if path.isfile(c)]

	@classmethod
	def get_example_stats_pairs(
		cls, ex_dir: str, stats_dir: str
	) -> Sequence[Tuple[str, str]]:
		"""
		Get a list of pairs of example file and its corresponding stats YAML
		file, based on the given root directories and in a recursive manner.

		Args:
			ex_dir: the root directory for the example files
			stats_dir: the root directory for the statistics YAML file
		
		Returns:
			A list of found pairs of example files and stats files
		"""
		ret = []
		for ex_path in cls.get_target_files(ex_dir):
			# check if it's an example file
			if cls.EXAMPLE_FILE_ROOT_CHECKER(ex_path):
				no_ext_path = path.splitext(ex_path)[0]
				stats_path = no_ext_path.replace(ex_dir, stats_dir) + '.yaml'
				# check if the corresponding stats file exists
				if path.exists(stats_path) and path.isfile(stats_path) and \
					cls.RESULT_FILE_ROOT_CHECKER(stats_path):
					ret.append((ex_path, stats_path))
		return ret

	@staticmethod
	def gen_random_line(min_len: int, max_len: int = 0) -> str:
		"""
		Generate some random lines.
		
		Args:
			min_len: the min number of lines if max_len is not 0, otherwise the max
			number of the lines
			max_len: the max number of lines if not 0
		
		Returns:
			the generated lines
		"""
		if not max_len:
			min_len, max_len = 0, min_len
		return ''.join(
			chr(randint(97, 120)) for _ in range(randint(min_len, max_len))
		)
	
	@staticmethod
	def prepare_analyser(script_path: str = None) -> MROAnalyser:
		"""Prepare a MROAnalyser populated with the content of the given script.

		Args:
			script_path: the path of the script to load by the analyser at start
		
		Returns:
			the prepared MROAnalyser
		"""
		analyser = MROAnalyser(TestUtils.EXAMPLE_FILE_ROOT)
		if script_path:
			analyser.replace_script_content(script_path, open(script_path).read())
		return analyser

	@classmethod
	def prepare_calculator(cls, script_path: str = None) -> MROCalculator:
		"""
		Prepare a MROCalculator populated with the content of the given script.

		Args:
			script_path: the path of the script to load by the analyser at start
		
		Returns:
			the prepared MROCalculator
		"""
		return cls.prepare_analyser(script_path).calculator
	
	@staticmethod
	def compare_code_lenses(actual_lenses, expected_lenses):
		"""
		Compare if the actual code lenses equal to the expected code lenses.
		"""
		assert len(actual_lenses) == len(expected_lenses)
		for actual_lens in actual_lenses:
			found = False
			for expected_lens in expected_lenses:
				if expected_lens['mro'] == actual_lens['data']:
					found = True
			assert found

	@staticmethod
	def assert_locations_in_ranges(
		ranges: List[Tuple[Tuple[int, int], Tuple[int, int]]],
		locations: List[Tuple[int, int]],
	):
		"""
		Check the locations are in the correspondent range.

		The number of locations should equal to the number of ranges.

		Args:
			ranges: the list of ranges corresponding to the locations
			locations: the list of locations to check
		"""
		# sort the ranges and locations
		ranges.sort()
		locations.sort()
		# the numbers should match
		assert len(ranges) == len(locations)
		# the location should be within the corresponding range
		for ran, loc in zip(ranges, locations):
			assert ran[0] <= loc < ran[1]


EX_YAML_PAIRS = TestUtils.get_example_stats_pairs(
	TestUtils.EXAMPLE_FILE_ROOT, TestUtils.RESULT_FILE_ROOT
)
"""The list of all the pairs for example file and its stats yaml file."""


@pytest.fixture
def load_yaml(yaml_path: str) -> Generator[dict, None, None]:
	"""The test fixture to load a yaml file.

	Args:
		yaml_path: the path to the yaml file

	Returns:
		the loaded yaml file in a dictionary
	"""
	with open(yaml_path, 'r') as yaml_file:
		yield yaml.load(yaml_file, Loader=yaml.Loader)


class TestTestUtils:
	"""Test suite for the test utility functions or variables."""

	@pytest.mark.parametrize(
		['ex_dir', 'stats_dir'],
		[
			(TestUtils.EXAMPLE_FILE_ROOT, TestUtils.RESULT_FILE_ROOT),
		]
	)
	def test_get_example_stats_pairs(self, ex_dir: str, stats_dir: str):
		"""Test if get_example_stats_pairs() can generate correct pair list."""
		pairs = TestUtils.get_example_stats_pairs(ex_dir, stats_dir)
		ex_paths = TestUtils.get_target_files(ex_dir)
		stats_paths = TestUtils.get_target_files(stats_dir)
		assert len(pairs) == len(ex_paths) == len(stats_paths)

	@pytest.mark.parametrize(
		['file_root', 'checker'],
		[
			[
				TestUtils.TEST_FILE_ROOT,
				TestUtils.TEST_FILE_ROOT_CHECKER,
			],
			[
				TestUtils.EXAMPLE_FILE_ROOT,
				TestUtils.EXAMPLE_FILE_ROOT_CHECKER,
			],
			[
				TestUtils.RESULT_FILE_ROOT,
				TestUtils.RESULT_FILE_ROOT_CHECKER,
			],
		]
	)
	def test_file_root(self, file_root: str, checker: Callable[[str], bool]):
		"""Test if a file root is containing the correct files."""
		candidate_set = TestUtils.get_target_files(file_root)
		assert len(candidate_set) > 0
		for file_path in candidate_set:
			assert checker(file_path), f'{file_path} has failed check...'

	@pytest.mark.parametrize('n_exec', [1000])
	def test_gen_random_line(self, n_exec):
		"""Test if gen_random_line() can generate a line of correct max length."""
		for _ in range(n_exec):
			max_len = randint(60, 100)
			assert len(TestUtils.gen_random_line(max_len)) <= max_len

	@pytest.mark.parametrize(
		['file_path', 'yaml_path'],
		[
			(yaml_path, yaml_path) for _, yaml_path in EX_YAML_PAIRS
		]
	)
	def test_load_yaml(self, load_yaml, file_path):
		"""Test if load_yaml() can correctly load a yaml file."""
		with open(file_path, 'r') as yaml_file:
			assert load_yaml == yaml.load(yaml_file, Loader=yaml.Loader)



"""The classic diamond problem for multiple inheritance.
"""
from typing import TypeVar, Generic


T = TypeVar('T')  # generic type for data


class A(Generic[T]):
    
    def func_a(self):
        pass

    def func_ab(self):
        pass

    def func_ac(self):
        pass

    def func_ad(self):
        pass


class B(A):
    
    def func_b(self):
        pass

    def func_bd(self):
        pass

    def func_ab(self):
        return super().func_ab()


class C(A):
    
    def func_c(self):
        pass

    def func_cd(self):
        pass

    def func_ac(self):
        return super().func_ac()


class D(B, C):
    
    def func_d(self):
        pass

    def func_ad(self):
        return super().func_ad()



from mrols.parsed_custom_class import ParsedCustomClass
from mrols.parsed_package_class import ParsedPackageClass, PARSED_OBJECT_CLASS
import jedi
from typing import Sequence, Dict, Set, List, Tuple
from jedi.api import Script
from jedi.api.classes import Name
from mrols.parsed_class import ParsedClass


class MROCalculator:
    """
    This class is responsible for calculating the MRO list of a given target and
    construct the code lens or hover response based on the calculated MRO list.
    It will also cache the intermediate results of the calculation to accelerate
    the potential similar requests in the future.
    """

    def __init__(
        self,
        root_dir: str,
    ) -> None:
        self.root_dir = root_dir
        self.project = jedi.Project(path=root_dir)
        # cache the actual codes as lines
        # this cache is very useful as there will be unsaved changes
        self.content_cache: Dict[str, Sequence[str]] = {}
        # script path -> Jedi script
        self.jedi_scripts_by_path: Dict[str, Script] = {}
        # script path -> ParsedClass list of the script
        self.parsed_names_by_path: Dict[str, Sequence[ParsedClass]] = {}
        # class full name -> ParsedClass
        self.parsed_name_by_full_name: Dict[str, ParsedClass] = {}
        # set of the outdated scripts' path
        self.outdated_scripts: Set[str] = set()

    def replace_content_in_cache(self, script_path: str, content: str) -> None:
        """
        To replace the cached content of a script by the new content.

        Args:
            script_path: the path of the target script
            content: the new content
        """
        lines = content.splitlines()
        # to add an empty line at the end if necessary as splitlines() will not
        # add an empty line if the last character is end of line `\n`
        if not content or content[-1] == '\n':
            lines.append('')
        # update the content cache
        self.content_cache[script_path] = lines
        # to mark the outdated script
        # this may lead to better performance in case of sequence of small
        # incremental changes
        self.mark_script_outdated(script_path)

    def update_content_in_cache(self, script_path: str, start_pos: Tuple[int,
                                                                         int],
                                end_pos: Tuple[int, int], change: str) -> None:
        """
        To update the cached content of a script by an incremental change.

        Args:
            script_path: the path of the target script
            start_pos: the start position (inclusive) of the changes, in format
                of (line, character)
            end_pos: the end position (exclusive) of the changes, in format of
                (line, character)
            change: the text of the incremental changes
        """
        # fetch the lines of the old content
        lines = self.content_cache[script_path]
        # decompose the start and end positions
        start_line, start_char = start_pos
        end_line, end_char = end_pos
        update_lines = change.splitlines()
        if not change or change[-1] == '\n':
            update_lines.append('')
        # the lines of the new content is consisted of three parts:
        new_lines = []  # placeholder for the new contents
        # 1. from the start of old content to the start position of the change
        new_lines.extend(lines[:start_line])
        new_lines.append(lines[start_line][:start_char])
        # 2. the change
        new_lines[-1] += (update_lines[0] if update_lines else '')
        new_lines.extend(update_lines[1:])
        # 3. from the end of the change to the end of the old content
        new_lines[-1] += lines[end_line][end_char:]
        new_lines.extend(lines[end_line + 1:])
        # update the content cache
        self.content_cache[script_path] = new_lines
        # to mark the outdated script
        self.mark_script_outdated(script_path)

    def _update_script(self, script_path: str):
        """
        To update the Jedi script and the ParsedClass list based on the given
        script path and its cached content.

        Args:
            script_path: the path of the target script
        """
        # remove the script out of the `outdated_scripts` to avoid infinite loop
        self.outdated_scripts.remove(script_path)
        # nothing to do if the script is not cached
        if script_path not in self.content_cache:
            return
        script = jedi.Script(
            code='\n'.join(self.content_cache[script_path]),
            path=script_path,
            project=self.project,
        )
        context = script.get_context()
        self.jedi_scripts_by_path[script_path] = script
        self.parsed_names_by_path[script_path] = [
            # ParsedClass.parse(class_name, script)
            self.parse_class_by_jedi_name(class_name)
            # ParsedClass.parse_by_jedi_name(class_name, self.jedi_scripts_by_path)
            for class_name in script.get_names()
            # only the class defined in the script will be considered
            if self._is_original_class(class_name, context)
        ]
        for parsed in self.parsed_names_by_path[script_path]:
            self.parsed_name_by_full_name[parsed.full_name] = parsed

    def mark_script_outdated(self, outdated_path: str):
        """
        Mark one script as outdated, so all relevant cached intermediate results
        are no longer valid and need to be recalculated when it's requested next
        time.

        Args:
            outdated_path: the path of the outdated script
        """
        if outdated_path in self.outdated_scripts:
            return
        self.outdated_scripts.add(outdated_path)
        if outdated_path in self.parsed_names_by_path:
            # delete all cached parsed classes
            for parsed in self.parsed_names_by_path[outdated_path]:
                self.parsed_name_by_full_name.pop(
                    parsed.full_name, None
                )
        # delete cached Jedi scripts and parsed classes entries
        self.jedi_scripts_by_path.pop(outdated_path, None)
        self.parsed_names_by_path.pop(outdated_path, None)

    def update_all(self):
        """
        Update all the outdated scripts.
        """
        # use a copy of the set as `self._update_script()`` will modify the set
        # in iteration
        for outdated_path in self.outdated_scripts.copy():
            self._update_script(outdated_path)

    def update_one(self, script_path: str):
        """
        Update the one specific outdated script, given its path.

        Args:
            script_path: the path of the target outdated script
        """
        if script_path in self.outdated_scripts:
            self._update_script(script_path)

    def get_code_lens(self, script_path: str) -> Sequence[Dict]:
        """
        Get the code lens list of the given script.

        Args:
            script_path: the path of the target script

        Returns:
            the list of code lens in the script
        """
        if script_path not in self.parsed_names_by_path:
            return []
        return [
            parsed.code_lens for parsed in self.parsed_names_by_path[script_path]
        ]

    def get_code_lens_and_range(
        self, script_path: str
    ) -> List[Tuple[Dict, Tuple[Tuple[int, int], Tuple[int, int]]]]:
        """
        Get the list of the code lens and the range of the associate parsed
        class for the given target script.

        Args:
            script_path: the path of the given script

        Returns:
            the list of the code lens and range
        """
        if script_path not in self.parsed_names_by_path:
            return []
        return [
            (
                parsed.code_lens,
                (
                    # changing to lines starting with 0
                    (parsed.start_pos[0] - 1, parsed.start_pos[1],),
                    (parsed.end_pos[0] - 1, parsed.end_pos[1],),
                )
            )
            for parsed in self.parsed_names_by_path[script_path]
        ]

    @staticmethod
    def _is_original_class(class_name: Name, script_context: Name) -> bool:
        """
        To check if a jedi Name is an originally defined class in a script.

        Args:
            class_name: the Name of the target class
            script_context: the context of the target script

        Returns:
            `True` if the class is an originally defined class or `False`
            otherwise
        """
        if not script_context.module_name:
            return class_name.type == 'class'
        # use `class_name.goto()[0]` to fetch the full content Jedi Name which
        # has the `full_name` field
        return class_name.type == 'class' and \
            class_name.goto()[0].full_name.startswith(
                script_context.module_name
            )

    def parse_class_by_jedi_name(
        self, jedi_name: Name
    ) -> ParsedClass:
        """
        To parse a class definition by its Jedi Name.

        Args:
            jedi_name: the Jedi Name of the target class to parse

        Returns:
            The parsed class in ParsedClass
        """
        # case of being the object class
        if jedi_name.full_name == 'builtins.object':
            return PARSED_OBJECT_CLASS
        # case of a class not defined by a recognised script, so external class
        if not jedi_name.module_path:
            return ParsedPackageClass(jedi_name)
        # case of a custom class definition, which should have a dot-separate
        # full name starting with the full name of the module/script, and its
        # type should be `class`. There is case that the first condition is
        # satisfied but the second not, like a type alias definition/assignment
        # has a type `statement`.
        # use `class_name.goto()[0]` to fetch the full content Jedi Name which
        # has the `full_name` field
        if jedi_name.goto()[0].full_name.startswith(
            jedi_name.module_name
        ) and jedi_name.type == 'class':
            return ParsedCustomClass(jedi_name, self)
        else:
            return ParsedPackageClass(jedi_name)



import jedi
from abc import ABC, abstractmethod
from typing import Tuple, Sequence, Dict
from jedi.api.classes import Name


class ParsedClass(ABC):
    """
    This class encapsulates a class definition parsed for MRO list calculation
    and all the intermediate results during the calculation.

    All the necessary calculation to get the MRO list will be done during the
    initialisation of the instance.
    """

    OBJECT_CLASS: Name = jedi.Script(code='object').infer(1, 0)[0]
    """A Jedi Name to represent the `object` class."""
    CONFLICT_MRO_MSG = 'Conflict MRO!!!'

    def __init__(self, jedi_name: Name) -> None:
        self.jedi_name = jedi_name
        self.full_name = self.jedi_name.full_name if self.jedi_name.full_name else ''
        self.start_pos: Tuple[int, int] = (0, 0)
        self.end_pos: Tuple[int, int] = (0, 0)
        self._code_lens = None

    @property
    @abstractmethod
    def mro_parsed_list(self) -> Sequence['ParsedClass']:
        """The MRO list in ParsedClass of the target class."""
        pass

    @property
    def code_lens(self) -> Dict:
        """The code lens correspondent to this parsed class."""
        if not self._code_lens:
            self._code_lens = self.get_code_lens()
        return self._code_lens

    @property
    def mro_list(self) -> Sequence[str]:
        """The MRO list of the class."""
        try:
            return [parsed.jedi_name.name for parsed in self.mro_parsed_list]
        except TypeError:
            return [self.CONFLICT_MRO_MSG]

    def get_code_lens(self):
        """Get the Code Lens associated with this parsed class."""
        return {
            'range': {
                'start': {
                    # changing to line starting with 0 (LSP standard)
                    'line': self.start_pos[0] - 1,
                    'character': self.start_pos[1],
                },
                'end': {
                    # changing to line starting with 0 (LSP standard)
                    'line': self.end_pos[0] - 1,
                    'character': self.end_pos[1] - 1,
                }
            },
            'data': self.mro_list,
        }

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, ParsedClass):
            return False
        return self.full_name == o.full_name



import pytest
from mrols.parsed_package_class import ParsedPackageClass, PARSED_OBJECT_CLASS
from tests.test_parsed_class import MockParsedClass


class TestParsedPackageClass:
	"""Test suite for the ParsedPackageClass."""

	@pytest.mark.parametrize('n_repeat', [100])
	def test_mro_parsed_list(self, n_repeat: int):
		for _ in range(n_repeat):
			random_jedi_name = MockParsedClass.random_jedi_name()
			package_class = ParsedPackageClass(random_jedi_name)
			assert package_class.mro_parsed_list == [package_class, PARSED_OBJECT_CLASS]
		assert ParsedPackageClass(ParsedPackageClass.OBJECT_CLASS) == PARSED_OBJECT_CLASS


import ast
import pytest
from os import path
from typing import Dict
from jedi.api import Script
from mrols.calculator import MROCalculator
from mrols.parsed_class import ParsedClass
from tests.test_utils import TestUtils, EX_YAML_PAIRS, load_yaml


class TestParsedCustomClass:
	"""Test suite for the ParsedCustomClass."""

	@staticmethod
	def prepare_calculator_and_script(script_path: str):
		# prepare the calculator
		calculator = TestUtils.prepare_calculator(script_path)
		# update the script in calculator
		calculator.update_all()
		# get the Jedi Script
		script = calculator.jedi_scripts_by_path[script_path]
		return calculator, script
	
	@staticmethod
	def construct_parsed_custom_class(
		calculator: MROCalculator, script: Script, expected: Dict
	):
		loc = expected['location']
		# get the Jedi Name, line index starting from 1
		class_name = script.infer(loc[0] + 1, loc[1])[0]
		# construct the ParsedCustomClass and return
		return ParsedCustomClass(class_name, calculator)

	@pytest.mark.parametrize(
		['script_path', 'yaml_path'],
		EX_YAML_PAIRS,
	)
	def test_get_code_lines(self, script_path, load_yaml):
		calculator, script = self.prepare_calculator_and_script(script_path)
		# get the original lines of the script
		script_lines = calculator.content_cache[script_path]
		# check against every custom defined class
		for expected in load_yaml.get('code_lenses', []):
			parsed_class = self.construct_parsed_custom_class(
				calculator, script, expected
			)
			parsed_lines = parsed_class._get_code_lines()
			# the beginning and the end may have extract codes, so using
			# `in` instead of `==`
			assert '\n'.join(parsed_lines) in '\n'.join(
				script_lines[
					expected['location'][0] : expected['location'][0] + len(parsed_lines)
				]
			)
		
	@pytest.mark.parametrize(
		['script_path', 'yaml_path'],
		EX_YAML_PAIRS,
	)
	def test_get_class_def_ast_from_lines(self, script_path, load_yaml):
		calculator, script = self.prepare_calculator_and_script(script_path)
		# check against every custom defined class
		for expected in load_yaml.get('code_lenses', []):
			parsed_class = self.construct_parsed_custom_class(
				calculator, script, expected
			)
			# check that there is one and only one class definition from the
			# code lines
			codes = '\n'.join(parsed_class._lines)
			mod = ast.parse(codes)
			assert len([n for n in mod.body if isinstance(n, ast.ClassDef)]) == 1
	
	@pytest.mark.parametrize(
		['script_path', 'yaml_path'],
		EX_YAML_PAIRS,
	)
	def test_mro_parsed_list(self, script_path, load_yaml):
		calculator, script = self.prepare_calculator_and_script(script_path)
		# check against every custom defined class
		for expected in load_yaml.get('code_lenses', []):
			parsed_class = self.construct_parsed_custom_class(
				calculator, script, expected
			)
			if expected['mro'] == [ParsedClass.CONFLICT_MRO_MSG]:
				with pytest.raises(TypeError):
					parsed_mro_list = parsed_class.mro_parsed_list
			else:
				# check the parsed mro list against the expected result
				parsed_mro_list = parsed_class.mro_parsed_list
				assert [p.jedi_name.name for p in parsed_mro_list] == expected['mro']


from mrols.parsed_custom_class import ParsedCustomClass



"""
This module provides examples for tests to check if the language server can
correctly deal with the case that there are syntax errors in the target script.

For example, there may be a good class declaration at the start of the script
followed by a bad class declaration. The language server should be able to
provide MRO information for the good one and no information for the bad one.
"""
from abc import ABC


class GoodClass(ABC):

	def __init__(self) -> None:
		self.val = 0


class BadClass(ABC)

	def __init__()
		self.val = 1


import keyword
import pytest
import jedi
from random import randint
from typing import Sequence, List
from jedi.api.classes import Name
from mrols.parsed_class import ParsedClass
from tests.test_utils import TestUtils


class MockParsedClass(ParsedClass):
	"""A mock implementation of ParsedClass for testing the latter's
	functionality."""

	def __init__(self, jedi_name: Name) -> None:
		super().__init__(jedi_name)
		# to store the Jedi Names of the MRO list
		self._mro_name_list : List[Name] = []
		# to store the ParsedClass of the MRO list
		self._mro_parsed_list : List[ParsedClass] = []
		# randomly generate the start and end positions
		self.start_pos = (randint(1, 10), randint(0, 80))
		self.end_pos = (self.start_pos[0] + randint(1, 10), randint(0, 80))

	@property
	def mro_parsed_list(self) -> Sequence[ParsedClass]:
		return self._mro_parsed_list
	
	@classmethod
	def random_mock_class(
		cls, n_mro: int = 5, max_name_len: int = 10
	) -> 'MockParsedClass':
		"""Randomly generate a MockParsedClass for test."""
		parsed = MockParsedClass(cls.random_jedi_name(max_name_len))
		# randomly generate the MRO list in Jedi Name
		parsed._mro_name_list = [
			cls.random_jedi_name(max_name_len) for _ in range(n_mro)
		]
		# construct the MRO list in ParsedClass from the one in Jedi Name
		parsed._mro_parsed_list = [
			MockParsedClass(n) for n in parsed._mro_name_list
		]
		return parsed

	@staticmethod
	def random_jedi_name(max_name_len: int = 10) -> Name:
		"""Randomly generate a Jedi Name for class definition."""
		class_name = TestUtils.gen_random_line(1, max_name_len)
		# regenerate the class name if the generated is a keyword
		while keyword.iskeyword(class_name):
			class_name = TestUtils.gen_random_line(1, max_name_len)
		code = f'class {class_name}: pass'
		return jedi.Script(code=code).infer(1, 6)[0]


class TestParsedClass:
	"""Test suite for the ParsedClass."""

	def test_object_class(self):
		assert ParsedClass.OBJECT_CLASS.full_name == 'builtins.object'

	@pytest.mark.parametrize('n_repeat', [100])
	def test_code_lens(self,  n_repeat: int):
		for _ in range(n_repeat):
			parsed = MockParsedClass.random_mock_class()
			# _code_lens should have not been calculated by default
			assert parsed._code_lens is None
			# calling the property to implicitly calculate the code lens
			fetched_code_lens = parsed.code_lens
			# check the correctness of the property
			assert parsed._code_lens == fetched_code_lens
	
	@pytest.mark.parametrize('n_repeat', [100])
	def test_mro_list(self, n_repeat: int):
		for _ in range(n_repeat):
			parsed = MockParsedClass.random_mock_class()
			assert parsed.mro_list == [n.name for n in parsed._mro_name_list]

	@pytest.mark.parametrize('n_repeat', [100])
	def test_get_code_lens(self, n_repeat: int):
		for _ in range(n_repeat):
			parsed = MockParsedClass.random_mock_class()
			lens = parsed.get_code_lens()
			# check the correctness of every fields
			assert lens['data'] == [n.name for n in parsed._mro_name_list]
			assert lens['range']['start']['line'] == parsed.start_pos[0] - 1
			assert lens['range']['start']['character'] == parsed.start_pos[1]
			assert lens['range']['end']['line'] == parsed.end_pos[0] - 1
			assert lens['range']['end']['character'] == parsed.end_pos[1] - 1

	@pytest.mark.parametrize('n_repeat', [100])
	def test_eq(self, n_repeat: int):
		for _ in range(n_repeat):
			# ParsedClass should not equal to non-ParsedClass
			assert MockParsedClass.random_mock_class() != 'text'
			assert MockParsedClass.random_mock_class() != 100
			# ParsedClass of same full name should equal
			common_name = MockParsedClass.random_jedi_name()
			assert MockParsedClass(common_name) == MockParsedClass(common_name)
			# ParsedClass of different full name should not equal
			other_name = MockParsedClass.random_jedi_name()
			while other_name.full_name == common_name.full_name:
				other_name = MockParsedClass.random_jedi_name()
			assert MockParsedClass(common_name) != MockParsedClass(other_name)
			# ParsedClass of same full name should equal, a second check
			random_parsed = MockParsedClass.random_mock_class()
			assert random_parsed == MockParsedClass(random_parsed.jedi_name)



import logging
import os
import socketserver
import threading
import argparse
from functools import partial
from typing import Callable, Optional
from mrols.mro_lang_server import MROLanguageServer


log = logging.getLogger(__name__)


class _StreamHandlerWrapper(socketserver.StreamRequestHandler, object):
    """A wrapper class that is used to construct a custom handler class."""

    delegate = None

    DELEGATE_CLASS: Optional[Callable] = None
    SHUTDOWN_CALL: Optional[Callable] = None

    def setup(self):
        super(_StreamHandlerWrapper, self).setup()
        # the DELEGATE_CLASS should be MROLanguageServer
        # pylint: disable=no-member
        self.delegate = self.DELEGATE_CLASS(self.rfile, self.wfile)

    def handle(self):
        try:
            self.delegate.start()
        except OSError as e:
            if os.name == 'nt':
                # Catch & pass on ConnectionResetError when parent process dies
                # pylint: disable=no-member, undefined-variable
                if isinstance(e, WindowsError) and e.winerror == 10054:
                    pass

        # pylint: disable=no-member
        self.SHUTDOWN_CALL()


def start_tcp_lang_server(bind_addr, port, check_parent_process, handler_class):
    if not issubclass(handler_class, MROLanguageServer):
        raise ValueError(
            'Handler class must be an instance of MROLanguageServer')

    def shutdown_server(check_parent_process, *args):
        # pylint: disable=unused-argument
        if check_parent_process:
            log.debug('Shutting down server')
            # Shutdown call must be done on a thread, to prevent deadlocks
            stop_thread = threading.Thread(target=server.shutdown)
            stop_thread.start()

    # Construct a custom wrapper class around the user's handler_class
    wrapper_class = type(
        handler_class.__name__ + 'Handler',
        (_StreamHandlerWrapper,),
        {'DELEGATE_CLASS': partial(handler_class,
                                   check_parent_process=check_parent_process),
         'SHUTDOWN_CALL': partial(shutdown_server, check_parent_process)}
    )

    server = socketserver.TCPServer((bind_addr, port),
                                    wrapper_class,
                                    bind_and_activate=False)
    server.allow_reuse_address = True

    try:
        server.server_bind()
        server.server_activate()
        log.info('Serving %s on (%s, %s)', handler_class.__name__, bind_addr,
                 port)
        server.serve_forever()
    finally:
        log.info('Shutting down')
        server.server_close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Start the Python MRO Language Server'
    )
    parser.add_argument(
        'port', default=3000, nargs='?', type=int,
        help='the port to use by the language server'
    )
    args = parser.parse_args()
    start_tcp_lang_server('127.0.0.1', args.port, False, MROLanguageServer)



import ast
from typing import Tuple, Sequence
from jedi.api.classes import Name
from mrols.parsed_class import ParsedClass
from mrols.parsed_package_class import ParsedPackageClass


class ParsedCustomClass(ParsedClass):
    """
    This class encapsulates a class definition parsed for MRO list calculation
    and all the intermediate results during the calculation.

    All the necessary calculation to get the MRO list will be done during the
    initialisation of the instance.
    """

    def __init__(self, jedi_name: Name, calculator: 'MROCalculator') -> None:  # type: ignore
        super().__init__(jedi_name)
        self._calculator = calculator
        # attention: resolve() will raise FileNotFoundError if in Python3.5-
        # or if pass strict=True in Python3.6+
        module_path = str(jedi_name.module_path.resolve())
        # the Jedi Name should have `module_path` available
        if (not module_path):
            raise ValueError(
                "Jedi Name's module name is not available or not registered.")
        # load the relevant script if not yet
        if module_path not in calculator.content_cache:
            with open(module_path) as loaded_module:
                calculator.replace_content_in_cache(
                    module_path, loaded_module.read())
        # update the analysis result of the relevant script if outdated
        if module_path in calculator.outdated_scripts:
            calculator.update_one(module_path)
        # the script should now be analysed and its path can be used to search
        if module_path not in calculator.jedi_scripts_by_path:
            raise ValueError(
                "Error when loading Jedi Name's module content into Calculator.")
        # fetch the relevant Jedi Script, get code lines and class definition
        self._jedi_script = calculator.jedi_scripts_by_path[module_path]
        self._lines = self._get_code_lines()
        self._class_def = self._get_class_def_ast_from_lines()
        # the Jedi Name should have the position information
        if self.jedi_name.line is None or self.jedi_name.column is None:
            raise ValueError(
                f'Parsed class {self.jedi_name.full_name} has no line or column information.')
        # positions with line starting with 1 (Jedi and AST standard)
        self.start_pos: Tuple[int, int] = (
            self.jedi_name.line,
            self.jedi_name.column
        )
        self.end_pos: Tuple[int, int] = (
            self.jedi_name.line,
            self.jedi_name.column + len(self._class_def.name)
        )
        # calculate the Jedi Names of the base parents
        self._base_parent_names = [
            self._jedi_script.infer(
                b.lineno + self.jedi_name.line - 1,
                b.col_offset
            )[0]
            for b in self._class_def.bases
        ] if self._class_def.bases else [self.OBJECT_CLASS]
        # refine the calculated Jedi Names of the base parents to full content
        for i, base in enumerate(self._base_parent_names):
            full_candidates = base.goto()
            if full_candidates and isinstance(full_candidates[0], Name):
                self._base_parent_names[i] = full_candidates[0]
        self._mro_parsed_list = None

    @property
    def mro_parsed_list(self) -> Sequence[ParsedClass]:
        if not self._mro_parsed_list:
            self._mro_parsed_list = self._get_mro_parsed_list()
        return self._mro_parsed_list

    def _get_code_lines(self):
        """Get the code block of the class definition, separated by lines."""
        # get the start and end positions and then the lines between (inclusive)
        start_pos = self.jedi_name.get_definition_start_position()
        end_pos = self.jedi_name.get_definition_end_position()
        lines = self.jedi_name.get_line_code(
            after=end_pos[0] - start_pos[0]
        ).splitlines()
        # trim the unwanted part in the first and the last lines
        # trim the last line first, otherwise the end position can be corrupted
        # when there is only one line
        lines[-1] = lines[-1][:end_pos[1]]
        lines[0] = lines[0][start_pos[1]:]
        return lines

    def _get_class_def_ast_from_lines(self):
        """Get the correspondent ast.ClassDef instance."""
        codes = '\n'.join(self._lines)
        mod = ast.parse(codes)
        # there will be one and only one class definition
        return [n for n in mod.body if isinstance(n, ast.ClassDef)][0]

    def _get_base_parent_parsed(self):
        """Get the list of base parent classes in ParseClass format."""
        return [
            self._calculator.parsed_name_by_full_name.get(
                base_name.full_name, ParsedCustomClass(
                    base_name, self._calculator)
            ) if base_name.full_name else ParsedPackageClass(base_name)
            for base_name in self._base_parent_names
        ]

    def _get_mro_parsed_list(self) -> Sequence[ParsedClass]:
        """
        Calculate the MRO list in ParsedClass via the C3 Linearisation
        algorithm.

        Details of the algorithm can be found in the `C3 Linearisation`
        Wikipedia page.
        """
        base_parent_parsed = self._get_base_parent_parsed()
        merge_list = [
            base_parsed.mro_parsed_list for base_parsed in base_parent_parsed]
        merge_list.append(base_parent_parsed)
        mro_parsed_list = [self] + self._merge_mro_parsed_lists(merge_list)
        return mro_parsed_list

    @classmethod
    def _merge_mro_parsed_lists(cls, sublists):
        """The merge step in the C3 Linearisation algorithm to merge the MRO
        sublists (elements in ParsedClass) to one result MRO list (elements in
        ParsedClass).
        """
        if not sublists:
            return []
        # iterate over all possible next parent class (the head)
        for i, mro_list in enumerate(sublists):
            head = mro_list[0]
            good_head = True
            # check the head candidate is not in any other list tail
            for cmp_list in sublists[:]:
                for parsed in cmp_list[1:]:
                    if head == parsed:
                        good_head = False
                        break
                if not good_head:
                    break
            # construct the remaining list to merge if the head is valid
            if good_head:
                next_list = []
                for merge_item in sublists:
                    new_list = [
                        item for item in merge_item
                        if item != head
                    ]
                    if new_list:
                        next_list.append(new_list)
                return [head] + cls._merge_mro_parsed_lists(next_list)
        raise TypeError('Cannot construct MRO list as conflict exists')



from typing import Dict, Sequence, Tuple, Optional
from mrols.calculator import MROCalculator


class MROAnalyser:
    """
    The actual analyser of the Python MRO Language Server. This analyser is
    responsible to manage the content of the target Python project, including
    both the persisted and unsaved contents, and to provide the analysis results
    for the MRO queries like the Hover requests and the CodeLens requests.

    All the line numbers and character/column numbers are 0-based according to
    the Language Server Protocol. Therefore, all the line numbers are
    immediately minused by one to change from 1-based (Jedi standard or
    AST standard) to 0-based (LSP standard).
    """

    def __init__(self, root_dir: str) -> None:
        # the root path of the Python project directory
        self.root_dir = root_dir
        # the MRO calculator responsible for all MRO relevant calculations
        self.calculator = MROCalculator(self.root_dir)

    def replace_script_content(self, script_path: str, content: str) -> None:
        """
        To replace the cached content of a script by the new content.

        Args:
            script_path: the path of the target script
            content: the new content
        """
        self.calculator.replace_content_in_cache(script_path, content)

    def update_script_content(self, script_path: str, start_pos: Tuple[int,
                                                                       int],
                              end_pos: Tuple[int, int], change: str) -> None:
        """
        To update the cached content of a script by an incremental change.

        Args:
            script_path: the path of the target script
            start_pos: the start position (inclusive) of the changes, in format
                of (line, character)
            end_pos: the end position (exclusive) of the changes, in format of
                (line, character)
            change: the text of the incremental changes
        """
        self.calculator.update_content_in_cache(
            script_path, start_pos, end_pos, change)

    def update_fetch_code_lens(self, script_path: str) -> Sequence[Dict]:
        """
        To update and fetch all MRO code lenses in the target script.

        Args:
            script_path: the path of the target script

        Returns:
            the list of the found MRO code lenses
        """
        self.calculator.update_one(script_path)
        return self.calculator.get_code_lens(script_path)

    def update_fetch_hover(self, script_path: str,
                           position: Tuple[int, int]) -> Optional[Dict]:
        """
        To update and fetch the hover information for a given position of a
        given script.

        Args:
            script_path: the path of the target script
            position: the position of the cursor to request hover information,
                in format (line, character)

        Returns:
            the hover information conforming to the Language Server Protocol
            Hover Request
        """
        self.calculator.update_one(script_path)
        for lens, (start_pos, end_pos) in \
                self.calculator.get_code_lens_and_range(script_path):
            if start_pos <= position < end_pos:
                return {
                    'contents': lens['data'],
                }



"""
This module provides example for tests to check if the language server can deal
with a complex MRO list where all the relevant class declarations are in the
same module here.

The example is taken from the [Wikipedia C3 linearization page](
https://en.wikipedia.org/wiki/C3_linearization). The original author keeps the
copyright and please refer to the wiki page for more details.
"""
class O: pass

class A(O): pass

class B(O): pass

class C(O): pass

class D(O): pass

class E(O): pass

class K1(A, B, C): pass

class K2(D, B, E): pass

class K3(D, A): pass

class Z(K1, K2, K3): pass


"""
This sub package includes multiple modules of test examples to check if the
language server can deal with a complex MRO list where all the relevant class
declarations are distributed in different files.

The example is taken from the [Wikipedia C3 linearization page](
https://en.wikipedia.org/wiki/C3_linearization). The original author keeps the
copyright and please refer to the wiki page for more details.
"""


from setuptools import setup
from os import path

DIR = path.dirname(path.abspath(__file__))
INSTALL_PACKAGES = open(path.join(DIR, 'requirements.txt')).read().splitlines()[1:]
TEST_PACKAGES = open(path.join(DIR, 'requirements-dev.txt')).read().splitlines()[1:]
TEST_PACKAGES = [pkg for pkg in TEST_PACKAGES if pkg not in INSTALL_PACKAGES]

with open(path.join(DIR, 'README.md')) as f:
    README = f.read()

setup(
    name='python-mro-language-server',
    packages=['mrols'],
    description="A simple Python language server to provide MRO (Method Resolution Order) inference via Hover and CodeLens.",
    long_description=README,
    long_description_content_type='text/markdown',
    install_requires=INSTALL_PACKAGES,
    version='0.0.1',
    url='https://github.com/mosckital/python-mro-language-server',
    author='Kaiyan XIAO',
    author_email='k.max.xiao@gmail.com',
    keywords=['MRO', 'language-server', 'method-resolution-order', 'syntax-analysis'],
    tests_require=TEST_PACKAGES,
    python_requires='>=3.6',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
    ],
)


from jedi.api.classes import Name
from mrols.parsed_class import ParsedClass


class ParsedPackageClass(ParsedClass):
    """
    This class encapsulates a class definition which is from an external package
    or the built-in packages, and any intermediate calculation results for the
    MRO list calculation.
    """

    def __init__(self, jedi_name: Name) -> None:
        super().__init__(jedi_name)
        self._mro_parsed_list = [self]
        if self.jedi_name.full_name != self.OBJECT_CLASS.full_name:
            self._mro_parsed_list.append(PARSED_OBJECT_CLASS)

    @property
    def mro_parsed_list(self):
        return self._mro_parsed_list


PARSED_OBJECT_CLASS = ParsedPackageClass(ParsedPackageClass.OBJECT_CLASS)
"""The parsed class for the built-in object class."""



"""
This module provides examples for tests to check if the language server can
correctly deal with the case that the MRO list of the target class declaration
has conflict due to bad inheritance tree.
"""
class O: pass

class A(O): pass

class B(A): pass

class C(A, B): pass


import pytest
from os import path
from random import randint
from mrols.analyser import MROAnalyser
from tests.test_utils import TestUtils, EX_YAML_PAIRS, load_yaml


class TestMROAnalyser:
    """Test suite for the MROAnalyser"""

    # region hover_tests
    @pytest.mark.parametrize(
        ['script_path', 'yaml_path'],
        EX_YAML_PAIRS,
    )
    def test_update_fetch_hover_success(self, script_path, load_yaml):
        """Test update_fetch_hover() against the successful test cases."""
        analyser = TestUtils.prepare_analyser(script_path)
        for lens in load_yaml.get('code_lenses', []):
            location, mro = lens['location'], lens['mro']
            hover = analyser.update_fetch_hover(
                script_path, (location[0], location[1])
            )
            assert hover and hover['contents'] == mro
    
    @pytest.mark.parametrize(
        ['script_path', 'yaml_path'],
        EX_YAML_PAIRS,
    )
    def test_update_fetch_hover_failure(self, script_path, load_yaml):
        """Test update_fetch_hover() against the failure test cases."""
        analyser = TestUtils.prepare_analyser(script_path)
        for failure in load_yaml.get('negative_cases', []):
            location = failure['location']
            hover = analyser.update_fetch_hover(
                script_path, (location[0], location[1])
            )
            assert hover is None
    # endregion hover_tests

    # region code_lens_tests
    @pytest.mark.parametrize(
        ['script_path', 'yaml_path'],
        EX_YAML_PAIRS,
    )
    def test_update_fetch_code_lens(self, script_path, load_yaml):
        """Test update_fetch_code_lens() against the successful test cases."""
        analyser = TestUtils.prepare_analyser(script_path)
        expected_lenses = load_yaml.get('code_lenses', [])
        # test code lens result with the original file content
        lenses = analyser.update_fetch_code_lens(script_path)
        TestUtils.compare_code_lenses(lenses, expected_lenses)
        # return if no need to test adding new content
        if 'dummy_content' not in load_yaml:
            return
        # add new test content into the file
        lines = analyser.calculator.content_cache[script_path]
        n_last_line = len(lines) - 1
        n_last_char = len(lines[-1])
        analyser.update_script_content(
            script_path,
            (n_last_line, n_last_char),
            (n_last_line, n_last_char),
            '\n'.join(load_yaml['dummy_content'])
        )
        # test code lens result with the new test content
        lenses = analyser.update_fetch_code_lens(script_path)
        expected_lenses.append(load_yaml['dummy_code_lens'])
        TestUtils.compare_code_lenses(lenses, expected_lenses)
    # endregion code_lens_tests

    # region content_update_functions
    @pytest.mark.parametrize(
        ('n_times', 'max_lines', 'max_line_len'),
        [(10, 50, 80)],
    )
    def test_replace_script_content(self, n_times: int, max_lines: int,
                                    max_line_len: int):
        """Test the correctness of the replace_script_content() method."""
        for _ in range(n_times):
            content = '\n'.join(
                TestUtils.gen_random_line(max_line_len)
                for _ in range(max_lines)
            )
            analyser = MROAnalyser('')
            file_path = 'test_file.py'
            analyser.replace_script_content(file_path, content)
            assert '\n'.join(analyser.calculator.content_cache[file_path]) == content

    @pytest.mark.parametrize(
        ('n_times', 'max_lines', 'max_line_len'),
        [(10, 50, 80)],
    )
    def test_update_script_content(self, n_times: int, max_lines: int,
                                    max_line_len: int):
        """Test the correctness of the update_script_content() method."""
        for _ in range(n_times):
            # init analyser
            analyser = MROAnalyser('')
            file_path = 'test_file.py'
            # generate original content
            original_lines = [
                TestUtils.gen_random_line(max_line_len)
                for _ in range(max_lines)
            ]
            original_content = '\n'.join(original_lines)
            analyser.replace_script_content(file_path, original_content)
            # prepare incremental change range info
            start_line = randint(0, len(original_lines) - 1)
            end_line = randint(0, len(original_lines) - 1)
            start_char = randint(0, len(original_lines[start_line]))
            end_char = randint(0, len(original_lines[end_line]))
            if (start_line, start_char) > (end_line, end_char):
                start_line, start_char = end_line, end_char
            # prepare incremental change content
            change_lines = [
                TestUtils.gen_random_line(max_line_len)
                for _ in range(max_lines)
            ]
            change_content = '\n'.join(change_lines)
            # get invariant part before/after the change content
            prev_invariant = original_lines[:start_line + 1]
            prev_invariant[-1] = prev_invariant[-1][:start_char]
            prev_invariant = '\n'.join(prev_invariant)
            post_invariant = original_lines[end_line:]
            post_invariant[0] = post_invariant[0][end_char:]
            post_invariant = '\n'.join(post_invariant)
            # combine parts into the new content
            new_content = ''.join(
                [prev_invariant, change_content, post_invariant])
            # apply incremental update and check assertion
            analyser.update_script_content(file_path, (start_line, start_char),
                                           (end_line, end_char),
                                           change_content)
            assert '\n'.join(analyser.calculator.content_cache[file_path]) == new_content
    # endregion content_update_functions



import pytest
from os import path
from mrols.calculator import MROCalculator
from tests.test_utils import TestUtils, EX_YAML_PAIRS, load_yaml


class TestMROCalculator:
	"""Test suite for the MROCalculator.
	
	The replace_content_in_cache() and update_content_in_cache() are tested in
	test_analyser/TestMROAnalyser by test_replace_script_content() and
	test_update_script_content() as MROAnalyser.replace_script_content() and
	MROAnalyser.update_script_content() are actually wrappers of the two target
	functions.
	"""

	@staticmethod
	def assert_script_updated(
		calculator: MROCalculator,
		script_path: str,
		load_yaml: dict,
	):
		# the script should now be parsed into a Jedi Script and the content
		# should match
		assert script_path in calculator.jedi_scripts_by_path
		script = calculator.jedi_scripts_by_path[script_path]
		assert script._code.strip() == open(script_path).read().strip()
		# the class definitions in the script should also be parsed and cached
		# in the two dictionaries for lookup by script path or class full name
		assert script_path in calculator.parsed_names_by_path
		names = calculator.parsed_names_by_path[script_path]
		TestUtils.compare_code_lenses(
			[name.code_lens for name in names],
			load_yaml.get('code_lenses', []),
		)
		for name in names:
			assert name.full_name in calculator.parsed_name_by_full_name
			assert name == calculator.parsed_name_by_full_name[name.full_name]
	
	@staticmethod
	def assert_script_outdated(
		calculator: MROCalculator,
		script_path: str,
	):
		# the script should NOT be parsed into a Jedi Script
		assert script_path not in calculator.jedi_scripts_by_path
		# there should be NO parsed classes relating to the script
		assert script_path not in calculator.parsed_names_by_path
		abs_script_path = path.abspath(script_path)
		for name in calculator.parsed_name_by_full_name.values():
			if name.jedi_name.module_path:
				assert path.abspath(
					name.jedi_name.module_path
				) != abs_script_path

	@pytest.mark.parametrize(
		['script_path', 'yaml_path'],
		EX_YAML_PAIRS,
	)
	def test_update_script(self, script_path, load_yaml):
		# prepare the calculator
		calculator = TestUtils.prepare_calculator(script_path)
		# the script should be already in the content cache, but not parsed into
		# a Jedi Script
		assert script_path in calculator.content_cache
		assert script_path not in calculator.jedi_scripts_by_path
		# update the script
		calculator._update_script(script_path)
		# the script should be updated now
		self.assert_script_updated(calculator, script_path, load_yaml)

	@pytest.mark.parametrize(
		['script_path', 'yaml_path'],
		EX_YAML_PAIRS,
	)
	def test_mark_script_outdated(self, script_path, load_yaml):
		# prepare the calculator
		calculator = TestUtils.prepare_calculator(script_path)
		# the script should be already in the content cache, but not updated, so
		# acting like outdated
		assert script_path in calculator.content_cache
		self.assert_script_outdated(calculator, script_path)
		# update the script
		calculator.update_one(script_path)
		# the script should now be parsed into a Jedi Script
		assert script_path in calculator.content_cache
		self.assert_script_updated(calculator, script_path, load_yaml)
		# mark the script outdated
		calculator.mark_script_outdated(script_path)
		# the script should be outdated again
		assert script_path in calculator.content_cache
		self.assert_script_outdated(calculator, script_path)

	@pytest.mark.parametrize('n_outdated', [100])
	def test_update_all(self, n_outdated: int):
		# prepare the calculator
		calculator = TestUtils.prepare_calculator()
		# inject fake script paths into outdated_scripts
		calculator.outdated_scripts.update(
			f'fake_path_{i}' for i in range(n_outdated)
		)
		assert len(calculator.outdated_scripts) == n_outdated
		# update all and check
		calculator.update_all()
		assert len(calculator.outdated_scripts) == 0
	
	@pytest.mark.parametrize('n_outdated', [100])
	def test_update_one(self, n_outdated: int):
		# the fake target script path
		target_script = 'target_script'
		# prepare the calculator
		calculator = TestUtils.prepare_calculator()
		# inject fake script paths into outdated_scripts
		calculator.outdated_scripts.update(
			f'fake_path_{i}' for i in range(n_outdated)
		)
		calculator.outdated_scripts.add(target_script)
		assert len(calculator.outdated_scripts) == n_outdated + 1
		# update the target one and check
		calculator.update_one(target_script)
		assert len(calculator.outdated_scripts) == n_outdated
	
	@pytest.mark.parametrize(
		['script_path', 'yaml_path'],
		EX_YAML_PAIRS,
	)
	def test_get_code_lens(self, script_path, load_yaml):
		# prepare the calculator
		calculator = TestUtils.prepare_calculator(script_path)
		# update and then check the code lenses are correct
		calculator.update_all()
		lenses = calculator.get_code_lens(script_path)
		TestUtils.compare_code_lenses(
			lenses,
			load_yaml.get('code_lenses', []),
		)
	
	@pytest.mark.parametrize(
		['script_path', 'yaml_path'],
		EX_YAML_PAIRS,
	)
	def test_get_code_lens_and_range(self, script_path, load_yaml):
		# prepare the calculator
		calculator = TestUtils.prepare_calculator(script_path)
		# update the script
		calculator.update_all()
		# fetch the code lenses and ranges and separate them
		fetched = calculator.get_code_lens_and_range(script_path)
		lenses = [t[0] for t in fetched]
		ranges = [t[1] for t in fetched]
		# check the code lenses are correct
		TestUtils.compare_code_lenses(
			lenses,
			load_yaml.get('code_lenses', []),
		)
		# check the ranges are correct
		locations = [
			(
				lens['location'][0], lens['location'][1]
			) for lens in load_yaml.get('code_lenses', [])
		]
		TestUtils.assert_locations_in_ranges(ranges, locations)

	@pytest.mark.parametrize(
		['script_path', 'yaml_path'],
		EX_YAML_PAIRS,
	)
	def test_is_original_class(self, script_path, load_yaml):
		# prepare the calculator
		calculator = TestUtils.prepare_calculator(script_path)
		# update the script
		calculator.update_all()
		# fetch the correspondent Jedi Script and its Jedi Context
		script = calculator.jedi_scripts_by_path[script_path]
		context = script.get_context()
		# every ParsedClass should be based on an original Jedi Name
		for name in calculator.parsed_names_by_path[script_path]:
			assert calculator._is_original_class(name.jedi_name, context)
		# the number of original name should equal to the number of code lenses
		n_original = 0
		for name in script.get_names():
			if calculator._is_original_class(name, context):
				n_original += 1
		assert n_original == len(load_yaml.get('code_lenses', []))
	
	@pytest.mark.parametrize(
		'script_path',
		[
			s for s, _ in EX_YAML_PAIRS
		],
	)
	def test_parse_class_by_jedi_name(self, script_path):
		# prepare the calculator
		calculator = TestUtils.prepare_calculator(script_path)
		# update the script
		calculator.update_all()
		# fetch the correspondent Jedi Script and its Jedi Context
		script = calculator.jedi_scripts_by_path[script_path]
		context = script.get_context()
		from mrols.parsed_package_class import ParsedPackageClass
		from mrols.parsed_custom_class import ParsedCustomClass
		for name in script.get_names():
			# ignore the names that are not about class declarations
			if name.type != 'class':
				continue
			parsed_class = calculator.parse_class_by_jedi_name(name)
			if calculator._is_original_class(name, context):
				assert isinstance(parsed_class, ParsedCustomClass)
			else:
				assert isinstance(parsed_class, ParsedPackageClass)
